#!/usr/bin/env python
# coding: utf-8

import json
import traceback
import importlib
import os
import time
import importlib.util
import sys
from inspect import signature
from intelliw.config.cfg_parser import load_config
from intelliw.datasets.datasets import DataSets, DataSourceReaderException
from intelliw.utils import exception
from intelliw.utils.util import prepare_algorithm_parameters, prepare_model_parameters, import_code, get_first_element
from intelliw.utils.check_algorithm_py import is_valid_algorithm
import intelliw.utils.message as message
from intelliw.core.report import RestReport
from intelliw.core.recorder import Recorder
from intelliw.core.linkserver import linkserver
from intelliw.utils.logger import _get_framework_logger, _get_algorithm_logger
from intelliw.config import config
from intelliw.utils.global_val import gl
from intelliw import functions
from intelliw.core.al_decorator import make_decorators_server


logger = _get_framework_logger()


class Const:
    # 文件
    algorithmInformation = "AlgorithmInformation"
    algorithm = "algorithm"
    algorithm_py = "algorithm.py"
    algorithm_yaml = "algorithm.yaml"
    model_yaml = "model.yaml"

    # 处理过程
    pre_predict = "pre-predict"
    post_predict = "post-predict"
    pre_train = "pre-train"
    post_train = "post-train"

    # 其他
    new_target_name = "new_target_data_unrepeatable_name"
    metadata = "metadata"
    model_type = "modelType"
    target_name = "targetName"
    target_col = "targetCol"
    target_code = "targetCode"


def get_model_yaml():
    return Const.model_yaml


class Pipeline:
    def __init__(self, reporter=None, perodic_interval=-1):
        self.reporter = reporter if isinstance(
            reporter, RestReport) else RestReport(reporter)
        self.alldata_transforms = None       # 自定义针对总体数据结构处理的  特征工程
        self.splitdata_transforms = None     # 自定义针对单条数据数据结构处理的  特征工程
        self.feature_transforms = None        # 内置针对总体数据结构处理的  特征工程
        self.perodic_interval = perodic_interval
        self.recorder = Recorder(self.reporter, self.perodic_interval <= 0)
        self.instance = None
        self.custom_router = []
        self.metadata = []
        self.linkserver = None
        self.alg_describe = ''
        gl.set("reporter", self.reporter)

    def perodic_callback_train(self):
        logger.debug("定时任务： 上报状态")
        self.recorder.report()

    def perodic_callback_infer(self):
        logger.debug("定时任务： 上报状态")
        self.recorder.report_infer()

    def _check_algorithm_parameters_format(self, alg):
        try:
            alg_ps = alg['parameters'] if 'parameters' in alg else None
            if alg_ps is not None:
                for _, v in enumerate(alg_ps):
                    _ = v['key'] if 'key' in v else None
                    option = v['option'] if 'option' in v else None
                    if option is None:
                        return False
            return True
        except:
            return False

    def _prepare(self, path):
        sys.path.append(path)
        py_file = os.path.join(path, Const.algorithm_py)
        spec = importlib.util.spec_from_file_location(Const.algorithm, py_file)
        self.module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.module)
        return self.module

    def _initialize(self, alg_file, model_file=None):
        try:
            cfg = load_config(alg_file)
            gl.set("alg_yaml_file", alg_file)

            # 路由配置获取
            self.custom_router = cfg[Const.algorithmInformation].get(
                'router', [])

            # 算法配置获取
            alg_ps = cfg[Const.algorithmInformation][Const.algorithm].get(
                'parameters', None)
            parameters, option = prepare_algorithm_parameters(alg_ps)

            # model.yaml
            if model_file is not None:
                model_cfg = load_config(model_file)
                self.model_cfg = model_cfg['Model']

                # 算法描述
                self.alg_describe = model_cfg['Model'].get('desc', '')

                # 算法元数据
                self.metadata = self.model_cfg.get(Const.metadata, {})

                # 算法类型（分类回归时间序列。。。）
                self.model_type = self.metadata.get(Const.model_type)
                if self.model_type is not None:
                    gl.set("model_type", self.model_type)

                gl.set("model_yaml_file", model_file)

                model_ps = self.model_cfg.get('parameters')

                # model and algorithm's parameters are all in this var 'parameters'
                parameters = prepare_model_parameters(
                    model_ps, parameters, option)

                # 从环境变量中加载PARAMETERS
                env_model_ps = os.environ.get('PARAMETERS')
                if env_model_ps:
                    env_model_ps = json.loads(env_model_ps)
                    parameters = prepare_model_parameters(
                        env_model_ps, parameters, option)

                # 特征处理可能会修改target_col，所以需要先保存起来, 有target_name时一定会有target_col
                target_metadata_cfg = self.metadata.get("target", [])
                target_metadata = None
                if len(target_metadata_cfg) > 0:
                    target_metadata = [{"target_col": int(m[Const.target_col]), "target_name":m.get(
                        Const.target_code, m[Const.target_name])} for m in target_metadata_cfg]
                else:
                    # TODO 暂时兼容 后面删了
                    target_col = parameters.get(Const.target_col)
                    target_name = parameters.get(Const.target_name)
                    if target_col is not None or target_name is not None:
                        target_col = int(
                            target_col) if target_col is not None else None
                        target_metadata = [
                            {"target_col": int(target_col), "target_name": target_name}]
                if target_metadata is not None:
                    gl.set("target_metadata", target_metadata)

                # 用户环境变量，取自metadata.userEnv
                user_env = self.metadata.get("userEnv", [])
                for ue in user_env:
                    key = str(ue.get("key", ""))
                    if key:
                        os.environ[key] = str(ue.get("value", ""))

                # linkserver
                self.linkserver = self.model_cfg.get('linkserver', None)
                linkserver.config = self.linkserver

            # 日志
            parameters["framework_log"] = _get_algorithm_logger()

            instance = self.module.Algorithm(parameters)
            if instance is None:
                self._raise_exception(message.err_import_alg_invalid_path)

            # 检测自定义路由方法是否存在
            for router in self.custom_router:
                func = router.get('func')
                if not hasattr(instance, func):
                    self._raise_exception(message.err_missing_router_func)

            # add decorator to  functions
            if config.is_server_mode():
                make_decorators_server(instance, self.reporter)

            return instance
        except Exception as e:
            self._raise_exception(message.err_import_alg_invalid_path, e)

    def _importalg_phase(self, path):
        if not os.path.isfile(os.path.join(path, Const.algorithm_py)):
            self._raise_exception(message.err_import_alg_missing_algorithpy)
        if not os.path.isfile(os.path.join(path, Const.algorithm_yaml)):
            self._raise_exception(message.err_import_alg_missing_algorithyaml)
        cfg = load_config(os.path.join(path, Const.algorithm_yaml))
        self.algorithm_cfg = cfg                        # SELF-DEFINE DICT
        if not self._check_algorithm_parameters_format(cfg):
            self._raise_exception(
                message.err_import_alg_invalid_algorithyaml_format)
        if self._prepare(path) is None:
            self._raise_exception(
                message.err_import_alg_invalid_algorithpy_format)
        is_valid, msg = is_valid_algorithm(self.module)
        if not is_valid:
            self._raise_exception(msg)

    def importalg(self, path):
        """
        导入算法
        :param path: 算法包所在路径
        :return: None
        :raises PipelineException: 框架导入算法失败原因
        """
        try:
            self._importalg_phase(path)
            self._initialize(os.path.join(path, Const.algorithm_yaml))
        except Exception as e:
            self._raise_exception(message.err_import_algorithm, e)
        self.reporter.report(message.ok_import_algorithm)

    def importmodel(self, path, load_model=True):
        """
        导入模型
        :param path: 算法包所在路径
        :param load_model: 是否要导入模型，即是否调用 Algorithm 的 load 方法
        :return: None
        :raises PipelineException: 框架导入模型失败原因
        """
        model_yaml_path = os.path.join(path, Const.model_yaml)
        algorithm_yaml_path = os.path.join(path, Const.algorithm_yaml)
        # check model yaml file
        if not os.path.isfile(model_yaml_path):
            logger.error(
                "未找到 model.yaml, path: {}, file_name: {}".format(path, Const.model_yaml))
            self._raise_exception(message.err_import_model_missing_modelyaml)

        try:
            # check algorithm package
            self._importalg_phase(path)

            self.instance = self._initialize(
                algorithm_yaml_path, model_yaml_path)

            # get transforms
            trans_cfg = self.model_cfg.get('transforms', None)
            transforms_result = self.prepare_transforms(trans_cfg)
            self.splitdata_transforms = transforms_result[0]
            self.alldata_transforms = transforms_result[1]
            self.feature_transforms = transforms_result[2]
            gl.set_dict({
                'feature_process': self.feature_transforms,
                'timeseries_process': transforms_result[3]
            })

            # load model
            if load_model:
                model_file = os.path.join(path, self.model_cfg['location'])
                self.instance.load(model_file)
        except Exception as e:
            self._raise_exception(message.err_import_model, e)

        # report
        self.reporter.report(message.ok_import_model)

    def _vaild_transforms_cfg(self, stage, is_split):
        transforms = self.splitdata_transforms if is_split else self.alldata_transforms
        if transforms is not None:
            transforms = transforms.get(stage)
            if transforms:
                return transforms
        return None

    def _transform_helper(self, stage, data, is_split):
        transforms = self._vaild_transforms_cfg(stage, is_split)
        if transforms is not None:
            return self.transform_call_chains(transforms, data)
        return data

    def _transform_function(self, stage, is_split):
        if self._vaild_transforms_cfg(stage, is_split) is None:
            return None

        def _transform(data):
            return self._transform_helper(stage, data, is_split)
        return _transform

    def _valid_feature_transforms(self, stage):
        if self.feature_transforms is not None:
            params = self.feature_transforms.get(stage)
            if params:
                return params
        return None

    def _feature_helper(self, stage, data):
        params = self._valid_feature_transforms(stage)
        if params is not None:
            from intelliw.functions.feature_process import feature_process
            return feature_process(data, params, stage)
        return data

    def _feature_process(self, stage):
        if self._valid_feature_transforms(stage) is None:
            return None

        def _process(data):
            return self._feature_helper(stage, data)
        return _process

    def _get_train_kwargs(self, reader, column_meta):
        if reader is None:
            return {}

        # config
        feature_process_cfg = gl.pop("feature_process")
        if feature_process_cfg is not None:
            feature_process_cfg = {
                Const.pre_train: feature_process_cfg.get(Const.pre_train),
                Const.post_train: feature_process_cfg.get(Const.post_train)
            }

        final_target_col = gl.pop("final_target_col")
        target_cols = gl.pop("target_cols")
        target_metadata = gl.pop("target_metadata")
        if target_cols is None and target_metadata is not None and len(target_metadata) > 0:
            target_cols = [
                {
                    "col": m["target_col"],
                    "column_name":m["target_name"]
                } for m in target_metadata
            ]
            final_target_col = target_cols[-1]["col"]

        kwargs = {
            "final_target_col": final_target_col,
            "target_cols": target_cols,
            "category_cols": gl.pop("category_cols"),
            "column_meta": gl.pop("column_meta") if gl.get("column_meta") is not None else [i["code"] for i in column_meta],
            "column_relation_df": gl.pop("column_relation_df"),
            "feature_process_cfg": feature_process_cfg,
        }

        # 时间序列参数
        if gl.get("model_type") == 9:
            timeseries_process = gl.pop("timeseries_process")
            if timeseries_process is not None:
                kwargs["timeseries_process_cfg"] = timeseries_process.get(
                    Const.pre_train)

        return kwargs

    def train(self, datasets: DataSets):
        if datasets is None:
            self._raise_exception(message.err_empty_train_data)

        # 特征值处理
        try:
            if config.SOURCE_TYPE == 0:
                train_set, valid_set, test_set = [], [], []
            else:
                train_set, valid_set, test_set = datasets.data_pipeline(
                    self._transform_function(Const.pre_train, True),
                    self._transform_function(Const.pre_train, False),
                    self._feature_process(Const.pre_train)
                )
        except Exception as e:
            self._raise_exception(message.err_function_process, e)

        # 开始训练
        try:
            logger.info("data process success, start train mode")

            # 数据集
            self.instance.train_set = train_set
            self.instance.valid_set = valid_set
            self.instance.test_set = test_set

            # 训练参数
            train_func = getattr(self.instance, "train")
            varnames = signature(train_func).parameters
            if "kwargs" not in varnames:
                # train(train, valid)
                train_func(
                    self.instance.train_set,
                    self.instance.valid_set
                )
            else:
                kwargs = self._get_train_kwargs(
                    train_set,
                    datasets.column_meta
                )
                if len(varnames) == 1:
                    # train(**kwargs)
                    train_func(**kwargs)
                elif len(varnames) == 3:
                    # train(train_set, valid_set, **kwargs)
                    train_func(
                        self.instance.train_set,
                        self.instance.valid_set,
                        **kwargs
                    )
                else:
                    self._raise_exception(message.err_missing_train_parameter)
            self.reporter.report(message.ok_train_finish)
        except exception.DatasetException as de:
            self._raise_exception(message.err_dataset, de,
                                  is_framework_error=False)
        except Exception as e:
            self._raise_exception(message.err_train, e,
                                  is_framework_error=False)

    @linkserver()
    def _infer_process(self, data, request, func='infer', need_feature=True):
        if need_feature:  # 前处理
            data = self._transform_helper(Const.pre_predict, data, False)
            data = self._feature_helper(Const.pre_predict, data)

        # 推理
        self.instance.request = request
        infer_func = getattr(self.instance, func)
        if infer_func.__code__.co_argcount > 1:  # 获取函数参数数量
            result = infer_func(data)
        else:
            result = infer_func()

        if need_feature:  # 后处理
            result = self._transform_helper(
                Const.post_predict, result, False)

        return result

    def infer(self, data, request, func='infer', need_feature=True):
        starttime = int(time.time() * 1000)
        try:
            logger.debug("receiving predict data {}".format(data))

            result = self._infer_process(data, request, func, need_feature)

            # 上报
            self.recorder.record_infer_status(
                '', True, starttime, int(time.time() * 1000))
            return result, None
        except Exception as e:
            stack_info = traceback.format_exc()
            logger.error("推理错误 {}, stack:\n{}".format(e, stack_info))
            self.recorder.record_infer_status('', False, starttime,
                                              int(time.time() * 1000), '{}'.format(e))
            return None, str("推理错误 {}".format(e))

    def transform_call_wrapper(self, src, func, params):
        if type(func) is type:
            return func().__call__(src, params)
        else:
            return func(src, params)

    def transform_call_chains(self, funcs, src):
        try:
            for _, item in enumerate(funcs):
                (name, func, params) = item
                fp = gl.get('feature_process')
                # 如果有特征工程前处理，就需要忽略select_columns， 要不会导致列下标错误
                if func == functions.select_columns and fp is not None and fp.get(Const.pre_train) is not None:
                    continue
                src = self.transform_call_wrapper(src, func, params)
            return src
        except Exception as e:
            msg = message.CommonResponse(
                500, "transforms", "特征工程函数 {} 出错".format(name))
            self._raise_exception(msg, e)

    def _get_func_by_name(self, name):
        """
        根据 name 加载函数

        加载顺序:
        1. 尝试在算法项目中加载
        2. 尝试在内置的特征处理中查找
        3. 尝试在当前 module 中加载

        :param name: 待加载的函数名称
        :return: 函数
        :raise: RuntimeError 加载失败原因
        """
        # 尝试在算法项目中加载
        mod = self.module
        if hasattr(mod, name):
            return getattr(mod, name)

        # 尝试在内置的特征处理中查找
        if hasattr(functions, name):
            return getattr(functions, name)

        # 尝试在当前 module 中加载
        mod = sys.modules[__name__]
        if hasattr(mod, name):
            return getattr(mod, name)

        raise RuntimeError('函数无效: [{}]'.format(name))

    def prepare_transforms(self, cfgs):
        def init_dict(f): return dict() if f is None else f
        def init_list(f): return list() if f is None else f
        # 没用到就不初始化，减少后面的判断
        split_transforms, alldate_transforms = None, None
        data_features, timeseries_features = None, None
        if cfgs:
            for v in cfgs:
                typ, functions_cfg = v.get('type'), v.get('functions')
                if typ is not None and functions_cfg is not None:
                    for item in functions_cfg:
                        key = item.get('key')
                        params, _ = prepare_algorithm_parameters(
                            item.get('parameters')
                        )
                        if key == functions.Const.feature_process:
                            # 特征工程
                            # 如果是离职预测，所有的特征auto配置必须为0，只做列数据筛选
                            data_features = init_dict(data_features)
                            data_features[typ] = json.loads(params.get(
                                functions.Const.feature_process_param, 'null')
                            )
                        elif key == functions.Const.timeseries_process:
                            # 时间序列
                            timeseries_features = init_dict(
                                timeseries_features)
                            timeseries_features[typ] = dict()

                            # normal
                            normal = json.loads(params.get(
                                functions.Const.timeseries_param, 'null')
                            )
                            if normal and normal.get("holiday", "none") == "none":
                                normal["holiday"] = None
                            timeseries_features[typ]["normal"] = normal

                            # group
                            group_params = json.loads(params.get(
                                functions.Const.timeseries_group_param, '[]')
                            )
                            for idx, i in enumerate(group_params):
                                k = f"industry_{idx}" if idx > 0 else "group"
                                timeseries_features[typ][k] = i
                        else:
                            # 常规数据处理
                            body = item.get('body')
                            func = self._get_func_by_name(
                                key) if body is None else get_first_element(import_code(body, None))
                            if func is not None:
                                can_split = key in functions.Const.split_function_list or item.get(
                                    'canSplit', False)

                                _t = split_transforms if can_split is True else alldate_transforms
                                _t = init_dict(_t)
                                _t[typ] = init_list(_t.get(typ))
                                _t[typ].append((key, func, params))
                                logger.info(
                                    "特征函数 {}, 是否可以进行分片处理 {}".format(key, can_split))
                            else:
                                logger.info("找不到函数 {}".format(key))
        return split_transforms, alldate_transforms, data_features, timeseries_features

    def validate_transforms(self, cfgs, data):
        """
        验证特征工程函数， 用于用户校验函数正确性和结果，服务于validateService
        """
        transforms = []
        if cfgs is not None:
            for _, v in enumerate(cfgs):
                functions = v['functions'] if 'functions' in v else None
                if functions is not None:
                    for k, item in enumerate(functions):
                        desc = item['desc'] if 'desc' in item else None
                        key = item['key'] if 'key' in item else None
                        body = item['body'] if 'body' in item else None
                        parameters = item['parameters'] if 'parameters' in item else None
                        if body is not None:
                            nmod = import_code(body, None)
                            func = get_first_element(nmod)
                        else:
                            func = self._get_func_by_name(key)
                        params, _ = prepare_algorithm_parameters(parameters)
                        if func is not None:
                            logger.info("特征函数 {}".format(key))
                            transforms.append((key, func, params))
                        else:
                            logger.info("找不到函数 {}".format(key))

        return self.transform_call_chains(transforms, data)

    def _raise_exception(self, msg: message.CommonResponse, e: Exception = None, is_framework_error=True):
        raise_logger = logger if is_framework_error else _get_algorithm_logger()

        # 用于传递错误给prepare_env.sh
        err_stack = "\n"
        if isinstance(e, (exception.DatasetException,
                          exception.DataSourceDownloadException,
                          exception.FeatureProcessException,
                          DataSourceReaderException,
                          ImportError)):
            err_stack += str(e)  # 数据集错误不需要调用栈，知道什么问题就行
        elif e:
            err_stack += f"✖️ {str(e)} ✖️\n{traceback.format_exc()}"
        else:
            err_stack = ""
        err_content = "[{}]\n{}".format(msg.msg, err_stack)
        os.environ["ERR_MASSAGE"] = err_content

        if e is None:
            raise_logger.error(msg.msg)
        else:
            msg.msg = err_content
            raise_logger.error(msg.msg + ' %s', e, exc_info=True)

        # 上报
        self.reporter.report(msg)
        sys.exit()
